import MathVectorHelper from '@/helpers/math/vector';

let HandlerManager = function (el) {
    this.handlers = [];
    this.el = el;
};
HandlerManager.prototype.add = function (handler) {
    this.handlers.push(handler);
};
HandlerManager.prototype.del = function (handler) {
    if (!handler) this.handlers = [];

    let delIndex = this.handlers.findIndex(h => h === handler);
    if (delIndex >= 0) this.handlers.splice(delIndex, 1);
};
HandlerManager.prototype.dispatch = function () {
    this.handlers.forEach(handler => {
        if (typeof handler === 'function') handler.apply(this.el, arguments);
    });
};

let wrapFunc = (el, handler) => {
    let manager = new HandlerManager(el);
    manager.add(handler);
    return manager;
};

let ScreenHandle = function (el, options) {
    this.id = Date.now();
    this.element = typeof el === 'string' ? document.querySelector(el) : el;
    this.options = Object.assign({
        offsetX: 0,
        offsetY: 0,
        clickTimeout: 150, // 点击触发时间 (ms)
        longclickTimeout: 750, // 长按触发时间 (ms)
        moveMomentum: 15, // 移动量
        pinchMomentum: 300, // 捏动量
        rotateMomentum: 20, // 转动量（度）
        bothPinchRotate: false // 是否可以同时捏和转
    }, options);
    this.onStart = this.onStart.bind(this);
    this.onMoving = this.onMoving.bind(this);
    this.onEnd = this.onEnd.bind(this);
    this.onCancel = this.onCancel.bind(this);
    this.usingMouse = false;
    if (window.ontouchstart === undefined) {
        this.usingMouse = true;
        this.element.addEventListener('mousedown', this.onStart, false);
        this.element.addEventListener('mousemove', this.onMoving, false);
        this.element.addEventListener('mouseup', this.onEnd, false);
        this.element.addEventListener('mouseleave', this.onCancel, false);
    } else {
        this.element.addEventListener('touchstart', this.onStart, false);
        this.element.addEventListener('touchmove', this.onMoving, false);
        this.element.addEventListener('touchend', this.onEnd, false);
        this.element.addEventListener('touchcancel', this.onCancel, false);
    }
    this.change = wrapFunc(this.element, options.onChange); // 事件变更（每一个事件）
    this.handle = wrapFunc(this.element, options.onHandle); // 操作事件
    // 基础事件
    this.start = wrapFunc(this.element, options.onStart); // 开始
    this.end = wrapFunc(this.element, options.onEnd); // 结束
    this.moving = wrapFunc(this.element, options.onMoving); // 移动
    this.cancel = wrapFunc(this.element, options.onCancel); // 中断
    // 点击事件
    this.singleClick = wrapFunc(this.element, options.onClick || options.onTap); // 单击
    this.doubleClick = wrapFunc(this.element, options.onDoubleClick || options.onDoubleTap); // 双击
    this.moreClick = wrapFunc(this.element, options.onMoreClick || options.onMoreTap); // 多击
    this.longClick = wrapFunc(this.element, options.onLongClick || options.onLongTap); // 长击
    // 移动事件
    this.swipe = wrapFunc(this.element, options.onSwipe); // 滑动
    this.swipeUp = wrapFunc(this.element, options.onSwipeUp); //
    this.swipeDown = wrapFunc(this.element, options.onSwipeDown);
    this.swipeLeft = wrapFunc(this.element, options.onSwipeLeft);
    this.swipeRight = wrapFunc(this.element, options.onSwipeRight);
    // 多指事件
    this.pinch = wrapFunc(this.element, options.onPinch);
    this.rotate = wrapFunc(this.element, options.onRotate);
    // 状态初始化
    this.init();
};
ScreenHandle.prototype = {
    init () {
        this.isMoving = false;
        this.points = [];
        this.clickcount = 0;
        this.clickTimeout = setTimeout(() => {});
        this.isLongClick = false;
        this.longclickTimeout = setTimeout(() => {});
    },
    computePointsDelta (from, to) {
        let offsetX = to.x - from.x;
        let offsetY = to.y - from.y;
        return Object.assign(to, {
            offset: {
                x: offsetX,
                y: offsetY,
                t: to.t - from.t
            }
        });
    },
    computePointsMathData (from, to) {
        let offsetX = to.x - from.x;
        let offsetY = to.y - from.y;
        return {
            directions: {
                n: offsetY < 0,
                s: offsetY > 0,
                w: offsetX < 0,
                e: offsetX > 0
            },
            direction: Math.abs(offsetX) >= Math.abs(offsetY) ? (offsetX < 0 ? 'Left' : 'Right') : (offsetY < 0 ? 'Up' : 'Down'),
            angle: MathVectorHelper.getPointsAngle(from, to),
            distance: MathVectorHelper.getDistance(from, to)
        };
    },
    setPoints (evt) {
        let points = [];
        if (this.usingMouse) {
            points.push({ x: evt.pageX + this.options.offsetX, y: evt.pageY + this.options.offsetY, t: Date.now() });
        } else {
            for (let i = 0; i < evt.touches.length; i++) {
                points.push({ x: evt.touches[i].pageX + this.options.offsetX, y: evt.touches[i].pageY + this.options.offsetY, t: Date.now() });
            };
        }
        points.forEach((point, index) => {
            if (!this.points[index]) {
                this.points.push({ from: point, to: this.computePointsDelta(point, point) });
            } else {
                let from = this.points[index].from;
                point = Object.assign(this.computePointsDelta(from, point));
                this.points[index] = Object.assign(this.points[index], this.computePointsMathData(from, point));
                this.points[index].to = point;
            }
        });
    },
    onStart (evt) {
        evt.preventDefault();
        this.setPoints(evt);
        if (this.points.length > 1) {
            // 多点
            clearTimeout(this.longclickTimeout);
        } else {
            // 单点
            this.longclickTimeout = setTimeout(() => {
                this.trigger('longClick', evt);
                this.isLongClick = true;
            }, this.options.longclickTimeout);
            this.clickcount++;
            if (this.points.length) {
                // 判断第二次点击与第一次的事件间隔，决定是否取消之前的单机事件。
                if (this.points[0].to.offset.t < this.options.clickTimeout) {
                    clearTimeout(this.clickTimeout);
                }
            }
        }
        this.isMoveable = true;
        this.trigger('start', evt);
    },
    onMoving (evt) {
        evt.preventDefault();
        if (!this.isMoveable) return;

        this.setPoints(evt);
        if (!this.isMoving) {
            let movingPoints = this.points.filter(point => {
                let momentum = (Math.abs(point.to.offset.x) + Math.abs(point.to.offset.y)) / 2;
                return momentum > this.options.moveMomentum;
            });
            if (movingPoints.length === 0) {
                return;
            }
        }

        clearTimeout(this.longclickTimeout);
        this.isMoving = true;
        this.trigger('moving', evt);

        if (this.points.length === 2) {
            let p1 = this.points[0];
            let p2 = this.points[1];
            let startDistance = MathVectorHelper.getDistance(p1.from, p2.from);
            let currentDistance = MathVectorHelper.getDistance(p1.to, p2.to);
            let rotate = p2.angle - p1.angle;
            if ((this.options.bothPinchRotate || !this.isRotate) && /* 判断是否可以发事件 */
                /* 判断是否是捏行为 */
                Math.abs(startDistance - currentDistance) > 25 /* 判断捏动量 */
            ) {
                this.isPinch = true;
                let scale = currentDistance / startDistance;
                this.trigger('pinch', evt, { scale });
            }
            if ((this.options.bothPinchRotate || !this.isPinch) && /* 判断是否可以发事件 */
                (p1.directions.n === p2.directions.s || p1.directions.w === p2.directions.e) &&/* 判断是否是转行为 */
                Math.abs(180 - rotate) > this.options.rotateMomentum /* 判断转角度 */
            ) {
                this.isRotate = true;
                this.trigger('rotate', evt, { rotate });
            }
        }
    },
    onEnd (evt) {
        evt.preventDefault();
        clearTimeout(this.longclickTimeout);
        this.setPoints(evt);
        if (!this.isMoving) {
            if (!this.isLongClick) {
                this.clickTimeout = ((points, clickcount) => {
                    return setTimeout(() => {
                        points[0].clickcount = clickcount;
                        switch (clickcount) {
                            case 1:
                                this.trigger('singleClick', evt, { points });
                                break;
                            case 2:
                                this.trigger('doubleClick', evt, { points });
                                break;
                            default:
                                this.trigger('moreClick', evt, { points });
                        };
                    }, this.options.clickTimeout);
                })(this.points, this.clickcount);
            }
        } else if (this.points.length === 1) {
            this.trigger('swipe', evt);
            this.trigger('swipe' + this.points[0].direction, evt);
        }
        this.isMoving = false;
        this.isPinch = false;
        this.isRotate = false;
        this.isMoveable = false;
        this.isLongClick = false;
        this.points = [];
        // this.trigger('end', evt);
    },
    onCancel (evt) {
        evt.preventDefault();
        if (this.points.length === 0) return;
        this.trigger('cancel', evt);
    },
    trigger (name, evt, data) {
        if (!this[name] || !this[name].dispatch || typeof this[name].dispatch !== 'function') {
            return;
        }
        let info = Object.assign({ name, points: this.points }, data);
        this[name].dispatch(evt, info);
        this.change.dispatch(evt, info);
        if (['singleClick', 'doubleClick', 'moreClick', 'longClick', 'moving', 'swipe', 'swipeUp', 'swipeDown', 'swipeLeft', 'swipeRight', 'pinch', 'pinchOpen', 'pinchClose', 'rotate', 'rotateLeft', 'rotateRight'].indexOf(name) !== -1) {
            // 确定的操作事件
            this.handle.dispatch(evt, info);
        }
        if (['start', 'end', 'moving', 'swipe', 'longClick', 'pinch', 'pinchOpen', 'pinchClose', 'rotate', 'rotateLeft', 'rotateRight'].indexOf(name) === -1) {
            // 不初始化的事件
            this.trigger('end', evt, Object.assign(info, { name: name + '.end' }));
            this.init();
        }
    },
    on (eventName, handler) {
        if (typeof this[eventName] === 'function') {
            this[eventName].add(handler);
        }
    },
    off (eventName, handler) {
        if (typeof this[eventName] === 'function') {
            this[eventName].del(handler);
        }
    }
};

export default ScreenHandle;
